import pytest
import pytest_asyncio

from redis import Redis
from redis.asyncio import Redis as AsyncRedis


@pytest.fixture(scope="function")
def redis():
    return Redis(host="127.0.0.1", port=6379)


@pytest_asyncio.fixture(scope="function")
async def aredis():
    async with AsyncRedis(host="127.0.0.1", port=6379) as conn:
        yield conn



from abc import abstractmethod

from redis_lock.base import BaseLock
from redis_lock.exceptions import AcquireFailedError


class BaseAsyncLock(BaseLock):
    """Base Redis lock implementation for asyncio"""

    async def __aenter__(self):
        if await self.acquire():
            return self
        raise AcquireFailedError("Failed to acquire a lock.")

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.release()

    @abstractmethod
    async def acquire(self, *args, **kwargs):
        """Try to acquire a lock"""
        raise NotImplementedError(
            "The `acquire` method should be implemented!"
        )

    @abstractmethod
    async def release(self):
        """Release the owned lock"""
        raise NotImplementedError(
            "The `release` method should be implemented!"
        )



import threading
import time

import pytest
from redis import Redis

from redis_lock import RedisLock, RedisSpinLock
from redis_lock.exceptions import (
    AcquireFailedError,
    InvalidArgsError,
    LockNotOwnedError,
)


class TestLock:
    def test_lock_arguments(self, redis: Redis):
        RedisLock(redis, name="key", blocking_timeout=1)
        RedisLock(redis, name="key", blocking_timeout=1.0)
        with pytest.raises(InvalidArgsError):
            RedisLock(redis, name=None)
        with pytest.raises(InvalidArgsError):
            RedisLock(redis, name="")
        with pytest.raises(InvalidArgsError):
            RedisLock(redis, name="key", blocking_timeout=None)
        with pytest.raises(InvalidArgsError):
            RedisLock(redis, name="key", blocking_timeout=-10)

        lock = RedisLock(redis, name="key")
        assert isinstance(lock.name, str)
        assert isinstance(lock.token, bytes)
        assert isinstance(lock.channel_name, str)

    def test_acquire_and_release(self, redis: Redis):
        name = "test_acquire_and_release"
        lock = RedisLock(redis, name)
        assert lock.acquire()
        assert lock.release()

    def test_acquire_with_context_manager(self, redis: Redis):
        name = "test_acquire_with_context_manager"
        with RedisLock(redis, name):
            assert redis.get(name)
        assert redis.get(name) is None

    def test_acquire_failed_error(self, redis: Redis):
        name = "test_acquire_failed_error"
        with RedisLock(redis, name):
            with pytest.raises(AcquireFailedError):
                with RedisLock(redis, name, blocking_timeout=.5):
                    pass
        redis.delete(name)

    def test_release_non_owned_lock_error(self, redis: Redis):
        name = "test_release_non_owned_lock_error"
        lock = RedisLock(redis, name)
        assert lock.acquire()
        redis.set(name, lock.token + b"foo")
        with pytest.raises(LockNotOwnedError):
            lock.release()
        redis.delete(name)

    def test_blocking_timeout(self, redis: Redis):
        name = "test_blocking_timeout"
        blocking_timeout = 1.0
        with RedisLock(redis, name):
            lock = RedisLock(redis, name, blocking_timeout=blocking_timeout)
            current = time.time()
            assert not lock.acquire()
            assert current + blocking_timeout < time.time()

    def test_pubsub_message(self, redis: Redis):
        name = "test_pubsub_message"
        lock = RedisLock(redis, name)

        def try_lock():
            another_lock = RedisLock(redis, name, blocking_timeout=1)
            assert another_lock.acquire()
            assert another_lock.release()

        lock.acquire()
        t = threading.Thread(target=try_lock)
        current = time.time()
        t.start()
        time.sleep(.5)
        lock.release()
        t.join()
        assert .5 < time.time() - current < 1

    def test_spin_lock(self, redis: Redis):
        name = "test_spin_lock"
        with RedisSpinLock(redis, name):
            assert redis.get(name)
        assert redis.get(name) is None

        with RedisSpinLock(redis, name):
            lock = RedisSpinLock(redis, name)
            assert not lock.acquire(blocking=False)

        with RedisSpinLock(redis, name):
            with pytest.raises(AcquireFailedError):
                with RedisSpinLock(redis, name, blocking_timeout=1):
                    pass

        lock = RedisSpinLock(redis, name)
        assert lock.acquire()
        redis.set(name, lock.token + b"foo")
        with pytest.raises(LockNotOwnedError):
            lock.release()
        redis.delete(name)



class RedisLockError(Exception):
    """Base class for exceptions related to Redis locks"""


class InvalidArgsError(RedisLockError):
    pass


class LockNotOwnedError(RedisLockError):
    pass


class AcquireFailedError(RedisLockError):
    pass



import asyncio
import time

import pytest
from redis.asyncio import Redis

from redis_lock.asyncio import RedisLock
from redis_lock.exceptions import (
    AcquireFailedError,
    InvalidArgsError,
    LockNotOwnedError,
)


class TestAsyncLock:
    @pytest.mark.asyncio
    async def test_lock_arguments(self, aredis: Redis):
        RedisLock(aredis, name="key", blocking_timeout=1)
        RedisLock(aredis, name="key", blocking_timeout=1.0)
        with pytest.raises(InvalidArgsError):
            RedisLock(aredis, name=None)
        with pytest.raises(InvalidArgsError):
            RedisLock(aredis, name="")
        with pytest.raises(InvalidArgsError):
            RedisLock(aredis, name="key", blocking_timeout=None)
        with pytest.raises(InvalidArgsError):
            RedisLock(aredis, name="key", blocking_timeout=-10)

        lock = RedisLock(aredis, name="key")
        assert isinstance(lock.name, str)
        assert isinstance(lock.token, bytes)
        assert isinstance(lock.channel_name, str)

    @pytest.mark.asyncio
    async def test_acquire_and_release(self, aredis: Redis):
        name = "async_test_acquire_and_release"
        lock = RedisLock(aredis, name)
        assert await lock.acquire()
        assert await lock.release()

    @pytest.mark.asyncio
    async def test_acquire_with_context_manager(self, aredis: Redis):
        name = "async_test_acquire_with_context_manager"
        async with RedisLock(aredis, name):
            assert await aredis.get(name)
        assert await aredis.get(name) is None

    @pytest.mark.asyncio
    async def test_acquire_failed_error(self, aredis: Redis):
        name = "async_test_acquire_failed_error"
        async with RedisLock(aredis, name):
            with pytest.raises(AcquireFailedError):
                async with RedisLock(aredis, name, blocking_timeout=.5):
                    pass
        await aredis.delete(name)

    @pytest.mark.asyncio
    async def test_release_non_owned_lock_error(self, aredis: Redis):
        name = "async_test_release_non_owned_lock_error"
        lock = RedisLock(aredis, name)
        assert await lock.acquire()
        await aredis.set(name, lock.token + b"foo")
        with pytest.raises(LockNotOwnedError):
            await lock.release()
        await aredis.delete(name)

    @pytest.mark.asyncio
    async def test_blocking_timeout(self, aredis: Redis):
        name = "async_test_blocking_timeout"
        blocking_timeout = 1.0
        async with RedisLock(aredis, name):
            lock = RedisLock(aredis, name, blocking_timeout=blocking_timeout)
            current = time.time()
            assert not await lock.acquire()
            assert current + blocking_timeout < time.time()

    @pytest.mark.asyncio
    async def test_pubsub_message(self, aredis: Redis):
        name = "async_test_pubsub_message"
        lock = RedisLock(aredis, name)

        async def try_lock():
            another_lock = RedisLock(aredis, name, blocking_timeout=1)
            assert await another_lock.acquire()
            assert await another_lock.release()

        async def release_lock():
            await asyncio.sleep(.5)
            await lock.release()

        await lock.acquire()
        current = time.time()
        await asyncio.gather(
            asyncio.create_task(try_lock()),
            asyncio.create_task(release_lock()),
        )
        assert .5 < time.time() - current < 1
